#include <WiFi.h>
#include <time.h>
#include <vector>
#include <sstream>
#include <vector>

#include <PubSubClient.h>

#include <SPIFFS.h>
#define FileSys SPIFFS

#include <PNGdec.h>
PNG png;
#define MAX_IMAGE_WIDTH 480
#define DISPLAY_WIDTH 240

#include "CronAlarms.h"

// Update these with values suitable for your network.
const char* ssid = "SSID";
const char* password = "PASSWORD";
const char* mqtt_server = "192.168.1.1";

static const char *TAG = "MAIN"; // logging tag
unsigned long previousMillis = 0;

WiFiClient espClient;
PubSubClient client(espClient);
#define MSG_BUFFER_SIZE	(128)
char msg[MSG_BUFFER_SIZE];


uint8_t bin_pos[3] = { 0, 115, 165 };
uint8_t bin_sel = 0;

// Include the TFT library https://github.com/Bodmer/TFT_eSPI
#include "SPI.h"
#include <TFT_eSPI.h>              // Hardware-specific library
TFT_eSPI tft = TFT_eSPI();         // Invoke custom library
#define TFT_BL_PIN 5
File pngfile;

// Dimming parameters
#define DIMMING_INTERVAL         10    // Milliseconds between brightness updates
const int BRIGHTNESS_INCREMENT  = 3;   // Adjusted for smooth dimming

// Activity signal parameters
#define PWM_MAX 1023
#define ACTIVITY_SIGNAL_INTERVAL     6000  // 6 seconds
#define ACTIVITY_DIMMING_INTERVAL    20    // Milliseconds between brightness updates during activity signal
#define MAX_BRIGHTNESS               128
const int ACTIVITY_SIGNAL_MIN_BRIGHTNESS = MAX_BRIGHTNESS / 2; // 50% brightness
const int ACTIVITY_BRIGHTNESS_CHANGE     = 5;                  // Adjusted for activity signal dimming

// Variables
bool led_on = false;                         // Control variable for LED state
int led_brightness = 0;                      // Current brightness level (0-255)
unsigned long last_update_time = 0;          // Last time brightness was updated
unsigned long last_activity_signal_time = 0; // Last time activity signal occurred

// Gamma correction parameters
#define GAMMA           2.2
#define GAMMA_TABLE_SIZE 256
uint16_t gamma_table[GAMMA_TABLE_SIZE];

// LED state machine
enum LedState {
  LED_OFF,
  LED_DIMMING_UP,
  LED_ON,
  LED_ACTIVITY_SIGNAL_DIMMING_DOWN,
  LED_ACTIVITY_SIGNAL_DIMMING_UP,
  LED_DIMMING_DOWN
};

LedState led_state = LED_OFF;




void generateGammaTable() {
  for( int i = 0; i < GAMMA_TABLE_SIZE; i++ ) {
    float normalized = (float)i / 255.0;
    float corrected = pow(normalized, GAMMA);
    gamma_table[i] = (uint16_t)(corrected * PWM_MAX + 0.5); // Add 0.5 for rounding
  }
}

// Function to update LED brightness and handle state transitions
void updateLED() {
  unsigned long current_time = millis();

  switch (led_state) {
    case LED_OFF:
      if (led_on) {
        led_state = LED_DIMMING_UP;
        last_update_time = current_time;
      }
      break;

    case LED_DIMMING_UP:
      if( current_time - last_update_time >= DIMMING_INTERVAL ) {
        last_update_time = current_time;
        led_brightness += BRIGHTNESS_INCREMENT;
        if( led_brightness >= MAX_BRIGHTNESS ) {
          led_brightness = MAX_BRIGHTNESS;
          led_state = LED_ON;
          last_activity_signal_time = current_time;
        }
        ledcWrite(0, gamma_table[ led_brightness ] );
      }
      break;

    case LED_ON:
      if( !led_on ) {
        led_state = LED_DIMMING_DOWN;
        last_update_time = current_time;
      } else if( current_time - last_activity_signal_time >= ACTIVITY_SIGNAL_INTERVAL ) {
        led_state = LED_ACTIVITY_SIGNAL_DIMMING_DOWN;
        last_update_time = current_time;
      }
      break;

    case LED_ACTIVITY_SIGNAL_DIMMING_DOWN:
      if( current_time - last_update_time >= ACTIVITY_DIMMING_INTERVAL ) {
        last_update_time = current_time;
        led_brightness -= ACTIVITY_BRIGHTNESS_CHANGE;
        if( led_brightness <= ACTIVITY_SIGNAL_MIN_BRIGHTNESS ) {
          led_brightness = ACTIVITY_SIGNAL_MIN_BRIGHTNESS;
          led_state = LED_ACTIVITY_SIGNAL_DIMMING_UP;
        }
        ledcWrite( 0, gamma_table[ led_brightness ] );
      }
      break;

    case LED_ACTIVITY_SIGNAL_DIMMING_UP:
      if (current_time - last_update_time >= ACTIVITY_DIMMING_INTERVAL) {
        last_update_time = current_time;
        led_brightness += ACTIVITY_BRIGHTNESS_CHANGE;
        if (led_brightness >= MAX_BRIGHTNESS) {
          led_brightness = MAX_BRIGHTNESS;
          led_state = LED_ON;
          last_activity_signal_time = current_time;
        }
        ledcWrite(0, gamma_table[ led_brightness ] );
      }
      break;

    case LED_DIMMING_DOWN:
      if (current_time - last_update_time >= DIMMING_INTERVAL) {
        last_update_time = current_time;
        led_brightness -= BRIGHTNESS_INCREMENT;
        if (led_brightness <= 0) {
          led_brightness = 0;
          led_state = LED_OFF;
        }
        ledcWrite(0, led_brightness);
      }
      break;
  }
}

// Callback function to draw pixels to the display
void pngDraw(PNGDRAW *pDraw) {
  uint16_t lineBuffer[MAX_IMAGE_WIDTH];
  png.getLineAsRGB565(pDraw, lineBuffer, PNG_RGB565_BIG_ENDIAN, 0xffffffff);

  // Pull 240px from the line we want to show on the screen
  uint16_t croppedLineBuffer[240];
  memcpy( croppedLineBuffer, &lineBuffer[bin_pos[bin_sel]], DISPLAY_WIDTH * sizeof(u_int16_t) );


  tft.pushImage(0, pDraw->y, DISPLAY_WIDTH, 1, croppedLineBuffer);
}


void * pngOpen(const char *filename, int32_t *size) {
  ESP_LOGI( TAG, "Attempting to open %s", filename );
  pngfile = FileSys.open(filename, "r");
  *size = pngfile.size();
  return &pngfile;
}

void pngClose(void *handle) {
  File pngfile = *((File*)handle);
  if (pngfile) pngfile.close();
}

int32_t pngRead(PNGFILE *page, uint8_t *buffer, int32_t length) {
  if (!pngfile) return 0;
  page = page; // Avoid warning
  return pngfile.read(buffer, length);
}

int32_t pngSeek(PNGFILE *page, int32_t position) {
  if (!pngfile) return 0;
  page = page; // Avoid warning
  return pngfile.seek(position);
}

void renderImage( uint8_t imageID ) {
  if( imageID < 3 ) {
    bin_sel = imageID;
  } else {
    bin_sel = 0;
  }

  int16_t rc = png.open( "/allbins_240.png", pngOpen, pngClose, pngRead, pngSeek, pngDraw);
  if (rc == PNG_SUCCESS) {
    tft.startWrite();
    ESP_LOGI( TAG, "image specs: (%d x %d), %d bpp, pixel type: %d", png.getWidth(), png.getHeight(), png.getBpp(), png.getPixelType());
    uint32_t dt = millis();
    if( png.getWidth() > MAX_IMAGE_WIDTH ) {
      ESP_LOGE( TAG, "Image too wide for allocated line buffer size!" );
    }
    else {
      rc = png.decode(NULL, 0);
      png.close();
    }
    tft.endWrite();
    // How long did rendering take...
    ESP_LOGI( TAG, "%d ms", (millis()-dt) );
  }
}

void setup_wifi() {
  delay(10);
  // We start by connecting to a WiFi network
  ESP_LOGI( TAG, "Connecting to %s", ssid );

  WiFi.mode( WIFI_STA);
  // Enable auto-reconnect
  WiFi.setAutoReconnect(true);

  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  randomSeed(micros());

  ESP_LOGI( TAG, "WiFi connected, IP address %s", WiFi.localIP().toString() );
}

std::string extractProperty(const char* topic) {
    std::string topicStr(topic);
    std::stringstream ss(topicStr);
    std::string segment;
    std::vector<std::string> segments;

    // Split the string by '/'
    while (std::getline(ss, segment, '/')) {
        segments.push_back(segment);
    }

    // Assuming the property is always at the 4th index
    if (segments.size() > 3) {
        return segments[3];
    } else {
        return "";
    }
}

void mqtt_callback(char* topic, byte* payload, unsigned int length) {

  String message = "";
  for (int i = 0; i < length; i++) {
    message += (char)payload[i];
  }
  ESP_LOGI( TAG, "Mesage arrived [%s] %s", topic, message );

  std::string prop = extractProperty( topic );
  ESP_LOGI( TAG, "Display %s [%s]", prop.c_str(), message.c_str() );
  if( message == "OFF" ) {
    led_on = false;
    tft.fillScreen( TFT_BLACK );
    digitalWrite(TFT_BL_PIN, LOW);
  } else if( message == "ON") {
    if( prop == "garbage" ) {
      led_on = true;
      digitalWrite(TFT_BL_PIN, HIGH);
      renderImage( 0 );
    } else if( prop == "recycling"  ) {
      led_on = true;
      digitalWrite(TFT_BL_PIN, HIGH);
      renderImage( 1 );
    } else if( prop == "organic" ) {
      led_on = true;
      digitalWrite(TFT_BL_PIN, HIGH);
      renderImage( 2 );
    }
  }
}

void reconnect() {
  // Loop until we're reconnected
  while( !client.connected() ) {
    ESP_LOGI( TAG, "Attempting MQTT connection...");
    // Create a random client ID
    char clientID[23];
    uint64_t chipid;
    chipid = ESP.getEfuseMac();
    uint16_t chip = (uint16_t)(chipid >> 32);

    snprintf( clientID, 23, "CartAlert%04X%08X", chip, (uint32_t)chipid );

    // Attempt to connect
    if( client.connect( clientID ) ) {
      ESP_LOGI( TAG, "MQTT connected" );

      String topic = "homie/";
      topic += clientID;

      client.publish( String( topic + "/$homie" ).c_str(), "3.0", true );
      client.publish( String( topic + "/$name" ).c_str(), "CartAlert", true );
      client.publish( String( topic + "/$nodes" ).c_str(), "bins", true );
      client.publish( String( topic + "/$implementation" ).c_str(), "Thinksilicon.de", true );
      client.publish( String( topic + "/$state" ).c_str(), "init", true );
      //client.publish( String( topic + "/$" ).c_str(), "", true );
      client.publish( String( topic + "/bins/$name" ).c_str(), "Bin indicator", true );
      client.publish( String( topic + "/bins/$type" ).c_str(), "240x240 pixels", true );
      client.publish( String( topic + "/bins/$properties" ).c_str(), "garbage,recycling,organic", true );
      client.publish( String( topic + "/bins/garbage/$name" ).c_str(), "Garbage bin", true );
      client.publish( String( topic + "/bins/garbage/$datatype" ).c_str(), "enum", true );
      client.publish( String( topic + "/bins/garbage/$settable" ).c_str(), "true", true );
      client.publish( String( topic + "/bins/garbage/$format" ).c_str(), "ON,OFF", true );
      client.publish( String( topic + "/bins/garbage" ).c_str(), "OFF", true );
      client.publish( String( topic + "/bins/recycling/$name" ).c_str(), "Recycling bin", true );
      client.publish( String( topic + "/bins/recycling/$datatype" ).c_str(), "enum", true );
      client.publish( String( topic + "/bins/recycling/$settable" ).c_str(), "true", true );
      client.publish( String( topic + "/bins/recycling/$format" ).c_str(), "ON,OFF", true );
      client.publish( String( topic + "/bins/recycling" ).c_str(), "OFF", true );
      client.publish( String( topic + "/bins/organic/$name" ).c_str(), "Organic and Recycling bin", true );
      client.publish( String( topic + "/bins/organic/$datatype" ).c_str(), "enum", true );
      client.publish( String( topic + "/bins/organic/$settable" ).c_str(), "true", true );
      client.publish( String( topic + "/bins/organic/$format" ).c_str(), "ON,OFF", true );
      client.publish( String( topic + "/bins/organic" ).c_str(), "OFF", true );

      client.publish( String( topic + "/$state" ).c_str(), "ready" );

      client.subscribe( String( topic + "/#" ).c_str() );

    } else {
      ESP_LOGE( TAG, "MQTT failed, rc=%d try again in 5 seconds", client.state() );
      // Wait 5 seconds before retrying
      delay(5000);
    }
  }
}

void displayOff() {
  ESP_LOGI( TAG, "Turning off display" );
  digitalWrite( TFT_BL_PIN, LOW );
}

void displayOn() {
  // check if it's still a garbo day
  if( led_on ) {
    ESP_LOGI( TAG, "Turning display back on" );
    digitalWrite( TFT_BL_PIN, HIGH );
    renderImage( bin_sel );
  }
}

//====================================================================================
//                                    Setup
//====================================================================================
void setup() {
  Serial.begin(115200);
  Serial.println("\n\n Init CartAlert" );

  // Initialise FS
  if (!FileSys.begin()) {
    ESP_LOGE( TAG, "SPIFFS initialisation failed!" );
    while (1) yield(); // Stay here twiddling thumbs waiting
  }

  // Initialise the TFT
  pinMode(TFT_BL_PIN, OUTPUT);
  digitalWrite(TFT_BL_PIN, HIGH); // Turn display ON
  tft.begin();
  tft.fillScreen(TFT_BLACK);
  digitalWrite( TFT_BL_PIN, LOW);

  ledcSetup( 0, 1000, 10 );
  ledcAttachPin( 15, 0 );
  ledcWrite( 0, 0 );
  generateGammaTable();

  setup_wifi();
  configTzTime( "MST7MDT,M3.2.0,M11.1.0", "pool.ntp.org" );
  struct tm timeinfo;
  while( !getLocalTime( &timeinfo ) ) {
    ESP_LOGE( TAG, "Failed to obtain time. Retry in 5s." );
    delay(5000);
  }
  Serial.println(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");

  client.setServer(mqtt_server, 1883);
  client.setCallback(mqtt_callback);

  Cron.create( "0  5  0 * * *", displayOff, false ); // turn off at night (00:05)
  Cron.create( "0  0  6 * * *", displayOn, false );  // turn back on in the morning (06:00)
  Cron.create( "0 15 10 * * *", displayOff, false ); // turn off when garbo picked up (10:15)

  Serial.println("\r\nInitialisation done.");
}

//====================================================================================
//                                    Loop
//====================================================================================
void loop() {
  if( WiFi.status() != WL_CONNECTED ) {
      unsigned long currentMillis = millis();
      // Attempt reconnection at defined intervals
      if( currentMillis - previousMillis >= 10000 ) {
        previousMillis = currentMillis;
        ESP_LOGE( TAG, "Wi-Fi disconnected. Attempting reconnection..." );
        WiFi.disconnect();
        delay( 50 );
        WiFi.reconnect();
      }
  } else {
    if( !client.connected() ) {
      reconnect();
    }
    client.loop();

    // Update LED based on the current state and timing
    updateLED();

    Cron.delay();
  }
}


